import time

TIMESINCE_CHUNKS: tuple[tuple[str, int, int | None], ...] = (
    ("second", 1, 60),
    ("minute", 60, 60),
    ("hour", 60 * 60, 24),
    ("day", 60 * 60 * 24, 6),
    ("week", 60 * 60 * 24 * 7, 6),
    ("month", 60 * 60 * 24 * 30, 11),
    ("year", 60 * 60 * 24 * 365, None),
)


def format_timesince(timestamp: float) -> str:
    delta: float = time.time() - timestamp
    if delta < 20:
        return "a few seconds ago"
    for label, divider, max_value in TIMESINCE_CHUNKS:
        value: float = round(number=delta / divider)
        if max_value is not None and value <= max_value:
            break
    return f"{value} {label}{'s' if value > 1 else ''} ago"


def format_size(size: int) -> str:
    size_float = float(size)
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(size_float) < 1000.0:
            return f"{size_float:3.1f}{unit}"
        size_float /= 1000.0
    return f"{size_float:.1f}Y"



import os
from pathlib import Path
from typing import Self

from pydantic import BaseModel

from bmhub.utils import format_size, format_timesince

FILES_TO_IGNORE: list[str] = [".DS_Store"]


class ModelInfo(BaseModel):
    id: str
    path: Path
    size_on_disk: int
    num_files: int
    last_accessed_timestamp: float
    last_modified_timestamp: float

    @property
    def last_accessed_str(self) -> str:
        return format_timesince(timestamp=self.last_accessed_timestamp)

    @property
    def last_modified_str(self) -> str:
        return format_timesince(timestamp=self.last_modified_timestamp)

    @property
    def size_on_disk_str(self) -> str:
        return format_size(size=self.size_on_disk)

    @classmethod
    def from_local_path(cls, *, model_path: Path) -> Self:
        model_id: str = f"{model_path.parent.name}/{model_path.name}"
        size_on_disk: int = 0
        num_files: int = 0
        last_accessed_timestamp: float = 0
        last_modified_timestamp: float = 0
        for file_path in model_path.glob(pattern="**/*"):
            if not file_path.is_file():
                continue
            if file_path.name in FILES_TO_IGNORE:
                continue
            file_stats: os.stat_result = file_path.stat()
            size_on_disk += file_stats.st_size
            num_files += 1
            last_accessed_timestamp = max(last_accessed_timestamp, file_stats.st_atime)
            last_modified_timestamp = max(last_modified_timestamp, file_stats.st_mtime)
        if num_files == 0:
            model_stats: os.stat_result = model_path.stat()
            last_accessed_timestamp = model_stats.st_atime
            last_modified_timestamp = model_stats.st_mtime
        return cls(
            id=model_id,
            path=model_path,
            size_on_disk=size_on_disk,
            num_files=num_files,
            last_accessed_timestamp=last_accessed_timestamp,
            last_modified_timestamp=last_modified_timestamp,
        )



import shutil
from os import terminal_size
from pathlib import Path
from typing import Annotated, Any

import typer
from rich import box
from rich.console import Console
from rich.markup import escape
from rich.table import Table

from bmhub.backend import HubBackend
from bmhub.schema import ModelInfo
from bmhub.utils import format_size

app = typer.Typer(no_args_is_help=True, help="BigModel Hub CLI Toolkit")

BACKEND_ARGUMENT: Any = typer.Argument(help="Model hub backend", show_default=False)
MODEL_ID_ARGUMENT: Any = typer.Argument(help="Model ID", show_default=False)
MODEL_ID_PATTERN_ARGUMENT: Any = typer.Argument(
    help="Glob pattern to match model IDs", show_default=False
)
LOCAL_DIR_OPTION: Any = typer.Option(
    "--local-dir",
    help="The local model directory",
    show_default=False,
    exists=True,
    file_okay=False,
    dir_okay=True,
    resolve_path=True,
)
YES_OPTION: Any = typer.Option(
    "--yes", "-y", help="Confirm the action without prompting", show_default=False
)


def get_console() -> Console:
    size: terminal_size = shutil.get_terminal_size()
    return Console(width=size.columns)


def print_model_table(console: Console, models: list[ModelInfo]) -> None:
    table = Table(box=box.SIMPLE)
    table.add_column(header="MODEL ID", no_wrap=True)
    table.add_column(header="SIZE ON DISK", justify="right", no_wrap=True)
    table.add_column(header="NB FILES", justify="right", no_wrap=True)
    table.add_column(header="LAST ACCESSED", no_wrap=True)
    table.add_column(header="LAST MODIFIED", no_wrap=True)
    table.add_column(header="MODEL PATH", no_wrap=True)
    for model in models:
        table.add_row(
            model.id,
            model.size_on_disk_str,
            str(object=model.num_files),
            model.last_accessed_str,
            model.last_modified_str,
            str(object=model.path),
        )
    num_models: int = len(models)
    size_on_disk_str: str = format_size(
        size=sum(model.size_on_disk for model in models)
    )
    console.print(table)
    console.print(f"Scanned {num_models} repo(s) for a total of {size_on_disk_str}.")


@app.command(name="list")
def list_(
    backend: Annotated[HubBackend, BACKEND_ARGUMENT],
    pattern: Annotated[str | None, MODEL_ID_PATTERN_ARGUMENT] = None,
    *,
    local_dir: Annotated[Path | None, LOCAL_DIR_OPTION] = None,
) -> None:
    """List downloaded models."""
    models: list[ModelInfo] = backend.list_models(pattern=pattern, local_dir=local_dir)
    console: Console = get_console()
    print_model_table(console=console, models=models)


@app.command()
def download(
    backend: Annotated[HubBackend, BACKEND_ARGUMENT],
    model_id: Annotated[str, MODEL_ID_ARGUMENT],
    *,
    local_dir: Annotated[Path | None, LOCAL_DIR_OPTION] = None,
) -> None:
    """Download a model from the model hub."""
    model_path: Path | None = None if local_dir is None else local_dir / model_id
    model_path_str: str = "cache" if model_path is None else f"<{model_path}>"
    console: Console = get_console()
    console.print(f"Downloading <{model_id}> to {model_path_str}...")
    backend.download_model(model_id=model_id, model_path=model_path)


@app.command()
def update(
    backend: Annotated[HubBackend, BACKEND_ARGUMENT],
    pattern: Annotated[str | None, MODEL_ID_PATTERN_ARGUMENT] = None,
    *,
    local_dir: Annotated[Path | None, LOCAL_DIR_OPTION] = None,
    yes: Annotated[bool, YES_OPTION] = False,
) -> None:
    """Update downloaded models."""
    models: list[ModelInfo] = backend.list_models(pattern=pattern, local_dir=local_dir)
    count: int = len(models)
    console: Console = get_console()
    console.print(f"The following {count} models will be updated:")
    print_model_table(console=console, models=models)
    console.print(
        escape(markup=f"Do you want to update the {count} models? [y/N] "), end=""
    )
    if not yes and not console.input().lower() == "y":
        return
    for i, m in enumerate(iterable=models):
        console.print(
            escape(markup=f"[{i + 1}/{count}] Updating <{m.id}> in <{m.path}>...")
        )
        backend.update_model(model=m)


@app.command()
def remove(
    backend: Annotated[HubBackend, BACKEND_ARGUMENT],
    pattern: Annotated[str | None, MODEL_ID_PATTERN_ARGUMENT] = None,
    *,
    local_dir: Annotated[Path | None, LOCAL_DIR_OPTION] = None,
    yes: Annotated[bool, YES_OPTION] = False,
) -> None:
    """Remove downloaded models."""
    models: list[ModelInfo] = backend.list_models(pattern=pattern, local_dir=local_dir)
    count: int = len(models)
    console: Console = get_console()
    console.print(f"The following {count} models will be removed:")
    print_model_table(console=console, models=models)
    console.print(
        escape(markup=f"Do you want to remove the {count} models? [y/N] "), end=""
    )
    if not yes and not console.input().lower() == "y":
        return
    for i, m in enumerate(iterable=models):
        console.print(
            escape(markup=f"[{i + 1}/{count}] Removing <{m.id}> from <{m.path}>...")
        )
        backend.remove_model(model=m)



import shutil
from fnmatch import fnmatch
from operator import attrgetter
from pathlib import Path

from modelscope.hub.snapshot_download import snapshot_download
from modelscope.utils.file_utils import get_model_cache_root

from bmhub.schema import ModelInfo


def get_ms_cache_dir() -> Path:
    cache_dir: Path = Path(get_model_cache_root()).expanduser().resolve()
    assert cache_dir.exists() and cache_dir.is_dir()
    return cache_dir


def _list_cached_ms_models() -> list[ModelInfo]:
    cache_dir: Path = Path(get_model_cache_root()).expanduser().resolve()
    if not cache_dir.exists() or cache_dir.is_file():
        raise ValueError(
            f"Cache directory not found: {cache_dir}. Please set `MODELSCOPE_CACHE` environment variable."  # noqa: E501
        )
    return _list_local_ms_models(local_dir=get_ms_cache_dir())


def _list_local_ms_models(*, local_dir: Path) -> list[ModelInfo]:
    models: list[ModelInfo] = []
    for organization_path in sorted(local_dir.iterdir()):
        if not organization_path.is_dir():
            continue
        if "___" in organization_path.name:
            continue
        for model_path in sorted(organization_path.iterdir()):
            if not model_path.is_dir():
                continue
            if "___" in model_path.name:
                continue
            if not (model_path / "config.json").exists():
                continue
            models.append(ModelInfo.from_local_path(model_path=model_path))
    return models


def list_ms_models(*, pattern: str | None, local_dir: Path | None) -> list[ModelInfo]:
    models: list[ModelInfo]
    if local_dir is None:
        models = _list_cached_ms_models()
    else:
        models = _list_local_ms_models(local_dir=local_dir)
    if pattern is not None:
        models = [model for model in models if fnmatch(name=model.id, pat=pattern)]
    return sorted(models, key=attrgetter("id"))


def download_ms_model(*, model_id: str, model_path: Path | None) -> None:
    if model_path is None:
        snapshot_download(repo_id=model_id)
        return
    snapshot_download(repo_id=model_id, local_dir=str(object=model_path))


def update_ms_model(*, model: ModelInfo) -> None:
    if model.path.is_relative_to(get_ms_cache_dir()):
        snapshot_download(repo_id=model.id)
        return
    snapshot_download(repo_id=model.id, local_dir=str(object=model.path))


def remove_ms_model(*, model: ModelInfo) -> None:
    if model.path.is_relative_to(get_ms_cache_dir()):
        if model.path.is_symlink():
            origin_path: Path = model.path.readlink()
            model.path.unlink()
            shutil.rmtree(origin_path)
        else:
            shutil.rmtree(model.path)
        return
    shutil.rmtree(model.path)



import shutil
from fnmatch import fnmatch
from operator import attrgetter
from pathlib import Path

from huggingface_hub import snapshot_download
from huggingface_hub.constants import HF_HUB_CACHE
from huggingface_hub.utils._cache_manager import (
    CachedRepoInfo,
    HFCacheInfo,
    scan_cache_dir,
)

from bmhub.schema import ModelInfo


def get_hf_cache_dir() -> Path:
    cache_dir: Path = Path(HF_HUB_CACHE).expanduser().resolve()
    assert cache_dir.exists() and cache_dir.is_dir()
    return cache_dir


def _list_cached_hf_models() -> list[ModelInfo]:
    return [
        ModelInfo(
            id=repo.repo_id,
            path=repo.repo_path,
            size_on_disk=repo.size_on_disk,
            num_files=repo.nb_files,
            last_accessed_timestamp=repo.last_accessed,
            last_modified_timestamp=repo.last_modified,
        )
        for repo in scan_cache_dir().repos
    ]


def _list_local_hf_models(*, local_dir: Path) -> list[ModelInfo]:
    models: list[ModelInfo] = []
    for organization_path in local_dir.iterdir():
        if not organization_path.is_dir():
            continue
        for model_path in organization_path.iterdir():
            if not model_path.is_dir():
                continue
            if not (model_path / "config.json").exists():
                continue
            models.append(ModelInfo.from_local_path(model_path=model_path))
    return models


def list_hf_models(*, pattern: str | None, local_dir: Path | None) -> list[ModelInfo]:
    models: list[ModelInfo]
    if local_dir is None:
        models = _list_cached_hf_models()
    else:
        models = _list_local_hf_models(local_dir=local_dir)
    if pattern is not None:
        models = [model for model in models if fnmatch(name=model.id, pat=pattern)]
    return sorted(models, key=attrgetter("id"))


def download_hf_model(*, model_id: str, model_path: Path | None) -> None:
    if model_path is None:
        snapshot_download(repo_id=model_id)
        return
    snapshot_download(repo_id=model_id, local_dir=model_path)


def update_hf_model(*, model: ModelInfo) -> None:
    if model.path.is_relative_to(get_hf_cache_dir()):
        snapshot_download(repo_id=model.id)
        return
    snapshot_download(repo_id=model.id, local_dir=model.path)


def remove_hf_model(*, model: ModelInfo) -> None:
    if model.path.is_relative_to(get_hf_cache_dir()):
        cache: HFCacheInfo = scan_cache_dir()
        repos: dict[str, CachedRepoInfo] = {repo.repo_id: repo for repo in cache.repos}
        assert model.id in repos
        revisions: list[str] = [
            revision.commit_hash for revision in repos[model.id].revisions
        ]
        cache.delete_revisions(*revisions).execute()
        return
    shutil.rmtree(model.path)



from enum import Enum
from pathlib import Path

from bmhub.hf_hub import (
    download_hf_model,
    get_hf_cache_dir,
    list_hf_models,
    remove_hf_model,
    update_hf_model,
)
from bmhub.ms_hub import (
    download_ms_model,
    get_ms_cache_dir,
    list_ms_models,
    remove_ms_model,
    update_ms_model,
)
from bmhub.schema import ModelInfo


class HubBackend(Enum):
    HUGGING_FACE = "hf"
    MODEL_SCOPE = "ms"

    def get_cache_dir(self) -> Path:
        match self:
            case HubBackend.HUGGING_FACE:
                return get_hf_cache_dir()
            case HubBackend.MODEL_SCOPE:
                return get_ms_cache_dir()

    def list_models(
        self, *, pattern: str | None, local_dir: Path | None
    ) -> list[ModelInfo]:
        match self:
            case HubBackend.HUGGING_FACE:
                return list_hf_models(pattern=pattern, local_dir=local_dir)
            case HubBackend.MODEL_SCOPE:
                return list_ms_models(pattern=pattern, local_dir=local_dir)

    def download_model(self, *, model_id: str, model_path: Path | None) -> None:
        match self:
            case HubBackend.HUGGING_FACE:
                return download_hf_model(model_id=model_id, model_path=model_path)
            case HubBackend.MODEL_SCOPE:
                return download_ms_model(model_id=model_id, model_path=model_path)

    def update_model(self, *, model: ModelInfo) -> None:
        match self:
            case HubBackend.HUGGING_FACE:
                return update_hf_model(model=model)
            case HubBackend.MODEL_SCOPE:
                return update_ms_model(model=model)

    def remove_model(self, *, model: ModelInfo) -> None:
        match self:
            case HubBackend.HUGGING_FACE:
                return remove_hf_model(model=model)
            case HubBackend.MODEL_SCOPE:
                return remove_ms_model(model=model)
